@@ -2,16 +2,20 @@ package Net::DNS::Resolver::Unbound;
 
 use strict;
 use warnings;
-
-our $VERSION;
-$VERSION = '1.19';
-
-use Carp;
 use Net::DNS;
 
-use base qw(Net::DNS::Resolver DynaLoader);
-eval { __PACKAGE__->bootstrap($VERSION) };
-warn "\n\n$@\n" if $@;
+use constant OS_SPEC => defined eval "require Net::DNS::Resolver::$^O";	## no critic
+use constant OS_CONF => join '::', 'Net::DNS::Resolver', OS_SPEC ? $^O : 'UNIX';
+use base qw(Net::DNS::Resolver::Base DynaLoader), OS_CONF;
+
+our $VERSION;
+
+BEGIN {
+	$VERSION = '1.22';
+	eval { __PACKAGE__->bootstrap($VERSION) };
+}
+
+use constant UB_SEND => Net::DNS::Resolver::Unbound::Context->can('ub_send');
 
 
 =head1 NAME
@@ -28,7 +32,8 @@ Net::DNS::Resolver::Unbound - Net::DNS resolver based on libunbound
 =head1 DESCRIPTION
 
 Net::DNS::Resolver::Unbound is designed as an extension to an existing
-Net::DNS installation which facilitates DNS(SEC) name resolution.
+Net::DNS installation which facilitates DNS(SEC) name resolution using
+the libunbound library developed by NLnet Labs.
 
 Net::DNS::Resolver::Unbound replaces the resolver send() and bgsend()
 functionality in the Net::DNS::Resolver::Base implementation.
@@ -75,12 +80,13 @@ Returns a new Net::DNS::Resolver::Unbound resolver object.
 =cut
 
 sub new {
-	my $class = shift;
-	my $self  = $class->SUPER::new();
+	my ( $class, %args ) = @_;
+	my @pass = map { ( $_, delete $args{$_} ) } qw(config_file);
+	my $self = $class->SUPER::new(@pass);
 	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new();
-	while ( my $attr = shift ) {
-		my $value = shift;
-		$self->$attr( ref($value) ? @$value : $value );
+	while ( my ( $attr, $value ) = each %args ) {
+		my $ref = ref($value);
+		$self->$attr( $ref ? @$value : $value );
 	}
 	return $self;
 }
@@ -93,12 +99,27 @@ sub new {
 	);
 
     my $fully_recursive = Net::DNS::Resolver::Unbound->new(
-	nameservers => [],		# override /etc/resolv.conf
+	nameservers => []		# override /etc/resolv.conf
 	);
 
+    my $dnssec_resolver = Net::DNS::Resolver::Unbound->new(
+	nameservers => [],
+	add_ta_file => '/var/lib/unbound/root.key'
+	);
+
+    my @nameservers = $stub_resolver->nameservers;
+
 By default, DNS queries are sent to the IP addresses listed in
 /etc/resolv.conf or similar platform-specific sources.
 
+=cut
+
+sub xnameservers {
+	my $self = shift;
+	local $self->{debug};		## "no nameservers" ok in this context
+	return $self->SUPER::nameservers(@_);
+}
+
 
 =head2 search, query, send, bgsend, bgbusy, bgread
 
@@ -107,24 +128,35 @@ See L<Net::DNS::Resolver>.
 =cut
 
 sub send {
-	my $self = shift;
+	my ( $self, @argument ) = @_;
 	$self->_finalise_config;
 	$self->_reset_errorstring;
 
-	my $query = $self->_make_query_packet(@_);
-	my ($q)	  = $query->question;
-	my $reply = $self->{ub_ctx}->ub_resolve( $q->name, $q->{qtype}, $q->{qclass} );
-	return $self->_decode_result( $reply, $query->header->id );
+	my ($packet) = @argument;
+	my $query = $self->_make_query_packet(@argument);
+	my $result;
+	if ( UB_SEND && ref($packet) ) {
+		$result = $self->{ub_ctx}->ub_send( $query->encode );
+	} else {
+		my ($q) = $query->question;
+		$result = $self->{ub_ctx}->ub_resolve( $q->name, $q->{qtype}, $q->{qclass} );
+	}
+
+	my $reply = $self->_decode_result($result);
+	$query->header->id( $reply->header->id );
+	return $reply;
 }
 
 sub bgsend {
-	my $self = shift;
+	my ( $self, @argument ) = @_;
 	$self->_finalise_config;
 	$self->_reset_errorstring;
 
-	my $query = $self->_make_query_packet(@_);
-	my ($q) = $query->question;
-	return $self->{ub_ctx}->ub_resolve_async( $q->name, $q->{qtype}, $q->{qclass}, $query->header->id );
+	my $query = $self->_make_query_packet(@argument);
+	my $image = $query->encode;
+	my $ident = $query->header->id;
+	my ($q)	  = $query->question;
+	return $self->{ub_ctx}->ub_resolve_async( $q->name, $q->{qtype}, $q->{qclass}, $ident );
 }
 
 sub bgbusy {
@@ -142,10 +174,10 @@ sub bgread {
 
 	$self->{ub_ctx}->ub_wait if &bgbusy;
 
-	my $async_id = $handle->async_id;
 	$self->errorstring( $handle->err );
-	my $result = $handle->result;
-	return $self->_decode_result( $result, $async_id );
+	my $reply = $self->_decode_result( $handle->result ) || return;
+	$reply->header->id( $handle->query_id );
+	return $reply;
 }
 
 
@@ -273,8 +305,8 @@ in RFC1035 zonefile format.
 =cut
 
 sub add_ta {
-	my $self = shift;
-	return $self->{ub_ctx}->add_ta( Net::DNS::RR->new(@_)->plain );
+	my ( $self, @argument ) = @_;
+	return $self->{ub_ctx}->add_ta( Net::DNS::RR->new(@argument)->plain );
 }
 
 
@@ -358,9 +390,10 @@ sub debug_level {
 
     $resolver->async_thread(1);
 
+Set the context behaviour for asynchronous actions.
 Enable a call to resolve_async() to create a thread to handle work in
-the background. If false (by default), a process is forked to perform
-the work.
+the background.
+If false (by default), a process is forked to perform the work.
 
 =cut
 
@@ -370,17 +403,17 @@ sub async_thread {
 }
 
 
-########################################
+=head2 print, string
 
-sub nameservers {
-	my $self = shift;
-	local $self->{debug};		## "no nameservers" ok in this context
-	return $self->SUPER::nameservers(@_);
-}
+    $resolver->print;
+    print $resolver->string;
+
+Prints the resolver state on the standard output.
+
+=cut
 
 sub string {
 	my $self = shift;
-
 	$self = $self->_defaults unless ref($self);
 
 	my @nslist   = $self->nameservers();
@@ -396,24 +429,21 @@ sub string {
 END
 }
 
-sub print {
-	return print shift->string;
-}
 
+########################################
 
 sub _decode_result {
-	my ( $self, $result, $query_id ) = @_;
+	my ( $self, $result ) = @_;
 
-	my $packet;
-	if ($result) {
-		$self->errorstring('INSECURE') unless $result->secure;
-		$self->errorstring( $result->why_bogus ) if $result->bogus;
+	return unless $result;
 
-		my $buffer = $result->answer_packet || return;
-		substr( $buffer, 0, 2 ) = pack 'n', $query_id;	# emulate Net::DNS resolver->send(packet)
-		$packet = Net::DNS::Packet->decode( \$buffer, $self->debug );
-		$self->errorstring($@);
-	}
+	$self->errorstring('INSECURE') unless $result->secure;
+	$self->errorstring( $result->why_bogus ) if $result->bogus;
+
+	my $buffer = $result->answer_packet || return;
+	my $packet = Net::DNS::Packet->decode( \$buffer );
+	$self->errorstring($@);
+	$packet->print if $self->debug;
 
 	return $packet;
 }
@@ -435,6 +465,7 @@ sub _finalise_config {
 	}
 
 	my $count = 3;
+	local $self->{debug};		## "no nameservers" ok in this context
 	foreach ( grep { $count-- > 0 } $self->nameservers ) {
 		$self->set_fwd($_);
 	}
@@ -448,7 +479,7 @@ __END__
 
 =head1 COPYRIGHT
 
-Copyright (c)2022 Dick Franks
+Copyright (c)2022,2023 Dick Franks
 
 All Rights Reserved
 
